{
open TestParser

exception Lexical_error of string


let int_literal_to_int ~base: base ~offset: offset ~lexem: lexeme =
  let lexeme_len = String.length lexeme in
  let int_value = ref Int32.zero in
  let base32 = Int32.of_int base in
  begin
  try
    for i = offset to lexeme_len -1 do
      let val_digit =
	(match lexeme.[i] with
	 |'0' .. '9' -> (Char.code lexeme.[i] - 48)
	 | 'a' .. 'f' -> 10 + (Char.code lexeme.[i] - 97)
	 | 'A' .. 'F' -> 10 + (Char.code lexeme.[i] - 65)
	 | 'u' | 'U' | 'l' | 'L' -> raise Exit
	 | _ -> assert false) in
      int_value :=
	(Int32.add (Int32.mul base32 !int_value) (Int32.of_int val_digit))
    done
  with Exit -> ()
  end ;
  !int_value
;;

}



let blank = [' ' '\009' '\012' '\026' ]
let pure_alpha = ['_' 'A'-'Z' 'a'-'z']
let identchar =
  ['A'-'Z' 'a'-'z' '_' '\192'-'\214' '\216'-'\246' '\248'-'\255' '\'' '0'-'9']
let is = ['u' 'U'  'l'  'L']
let decimal_literal = ['0'-'9']+ (is)*
let hex_literal = '0' ['x' 'X'] ['0'-'9' 'A'-'F' 'a'-'f']+ (is)*
let oct_literal = '0' ['o' 'O'] ['0'-'7']+ (is)*
let bin_literal = '0' ['b' 'B'] ['0'-'1']+ (is)*
    
rule token = parse
    blank +
    {
     token lexbuf
   }
  | ['\010' '\013']
      {
       incr Basics.line_counter ;
       token lexbuf
     }
  | '#' [^ '\010' '\013']* ('\010'|'\013')
      {
       incr Basics.line_counter ;
       token lexbuf
     }
  | '-' decimal_literal
      {
	let s = Lexing.lexeme lexbuf in
	NEG_INT_CONST (int_literal_to_int ~base: 10 ~offset: 1 ~lexem: s)
      }
  | decimal_literal
      {
	let s = Lexing.lexeme lexbuf in
	INT_CONST (int_literal_to_int ~base: 10 ~offset: 0 ~lexem: s)
     }
  | hex_literal
      {
       let s = Lexing.lexeme lexbuf in
       INT_CONST (int_literal_to_int ~base: 16 ~offset: 2 ~lexem: s)
     }
  | oct_literal
      {
       let s = Lexing.lexeme lexbuf in
       INT_CONST (int_literal_to_int ~base: 8 ~offset: 2 ~lexem: s)
     }
  | bin_literal
      {
       let s = Lexing.lexeme lexbuf in
       INT_CONST (int_literal_to_int ~base: 2 ~offset: 2 ~lexem: s)
     }
  | "%function"
      { FUNCTION_DIRECTIVE }
  | ':'
      { COLON }
  | '{'
      { LBRACE }
  | '}'
      { RBRACE }
  | "%parameters"
      { PARAMETERS_DIRECTIVE }
  | "%in"
      { IN_DIRECTIVE }
  | "%out"
      { OUT_DIRECTIVE }
  | "%inout"
      { INOUT_DIRECTIVE }
  | "INT"
      { INT_SPECIFICATION }
  | "STRING"
      { STRING_SPECIFICATION }
  | "TUPLE"
      { TUPLE_SPECIFICATION }
  | '['
      { LBRACKET }
  | ']'
      { RBRACKET }
  | '='
      { EQUALS }
  | "%globvars"
      { GLOBVARS_DIRECTIVE }
  | "%test"
      { TEST_DIRECTIVE }
  | "%inputs"
      { INPUTS_DIRECTIVE }
  | '('
      { LPAREN }
  | ')'
      { RPAREN }
  | ','
      { COMA  }
  | ';'
      { SEMI  }
  | "%results"
      { RESULTS_DIRECTIVE }
  | "\"" [^ '"']* "\""
      {
       let s = Lexing.lexeme lexbuf in
       QUOTED_STRING s
      }
  | pure_alpha identchar *
      {
       let s = Lexing.lexeme lexbuf in
       IDENTIFIER s
     }
  | eof
      { EOF }
  | _
      {
       let s = Lexing.lexeme lexbuf in
       raise (Lexical_error s)
     }

